@@ -21,6 +21,7 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdbool.h>
+#include <stdint.h>
 #include <inttypes.h>
 #include <string.h>
 #include <errno.h>
@@ -36,9 +37,11 @@
 #include <sys/vsock.h>
 #endif
 
-#include "internal.h"
+#include "array-size.h"
 #include "vector.h"
 
+#include "internal.h"
+
 #ifdef HAVE_LIBXML2
 
 #include <libxml/uri.h>
@@ -60,6 +63,13 @@ struct uri_query {
 
 DEFINE_VECTOR_TYPE (uri_query_list, struct uri_query);
 
+static void
+free_uri_query (struct uri_query q)
+{
+  free (q.name);
+  free (q.value);
+}
+
 /* Parse the query_raw substring of a URI into a list of decoded queries.
  * Return 0 on success or -1 on error.
  */
@@ -150,12 +160,38 @@ error:
   return -1;
 }
 
+/* Similar to nbdkit_parse_bool */
+static int
+parse_bool (const char *param, const char *value)
+{
+  if (!strcmp (value, "1") ||
+      !strcasecmp (value, "true") ||
+      !strcasecmp (value, "t") ||
+      !strcasecmp (value, "yes") ||
+      !strcasecmp (value, "y") ||
+      !strcasecmp (value, "on"))
+    return 1;
+
+  if (!strcmp (value, "0") ||
+      !strcasecmp (value, "false") ||
+      !strcasecmp (value, "f") ||
+      !strcasecmp (value, "no") ||
+      !strcasecmp (value, "n") ||
+      !strcasecmp (value, "off"))
+    return 0;
+
+  set_error (EINVAL, "could not parse %s parameter, expecting %s=true|false",
+             param, param);
+  return -1;
+}
+
 int
 nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
 {
   xmlURIPtr uri = NULL;
-  enum { tcp, unix_sock, vsock } transport;
-  bool tls, socket_required;
+  enum { tcp, unix_sock, vsock, ssh } transport;
+  bool tls;
+  enum { none, optional, required } socket_param;
   uri_query_list queries = empty_vector;
   int i, r;
   int ret = -1;
@@ -173,35 +209,45 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
 
   /* Scheme. */
   if (uri->scheme) {
-    if (strcmp (uri->scheme, "nbd") == 0) {
+    if (strcasecmp (uri->scheme, "nbd") == 0) {
       transport = tcp;
       tls = false;
-      socket_required = false;
+      socket_param = none;
     }
-    else if (strcmp (uri->scheme, "nbds") == 0) {
+    else if (strcasecmp (uri->scheme, "nbds") == 0) {
       transport = tcp;
       tls = true;
-      socket_required = false;
+      socket_param = none;
     }
-    else if (strcmp (uri->scheme, "nbd+unix") == 0) {
+    else if (strcasecmp (uri->scheme, "nbd+unix") == 0) {
       transport = unix_sock;
       tls = false;
-      socket_required = true;
+      socket_param = required;
     }
-    else if (strcmp (uri->scheme, "nbds+unix") == 0) {
+    else if (strcasecmp (uri->scheme, "nbds+unix") == 0) {
       transport = unix_sock;
       tls = true;
-      socket_required = true;
+      socket_param = required;
     }
-    else if (strcmp (uri->scheme, "nbd+vsock") == 0) {
+    else if (strcasecmp (uri->scheme, "nbd+vsock") == 0) {
       transport = vsock;
       tls = false;
-      socket_required = false;
+      socket_param = none;
     }
-    else if (strcmp (uri->scheme, "nbds+vsock") == 0) {
+    else if (strcasecmp (uri->scheme, "nbds+vsock") == 0) {
       transport = vsock;
       tls = true;
-      socket_required = false;
+      socket_param = none;
+    }
+    else if (strcasecmp (uri->scheme, "nbd+ssh") == 0) {
+      transport = ssh;
+      tls = false;
+      socket_param = optional;
+    }
+    else if (strcasecmp (uri->scheme, "nbds+ssh") == 0) {
+      transport = ssh;
+      tls = true;
+      socket_param = optional;
     }
     else {
       set_error (EINVAL, "unknown NBD URI scheme: %s", uri->scheme);
@@ -241,7 +287,9 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
       (transport == unix_sock &&
        (h->uri_allow_transports & LIBNBD_ALLOW_TRANSPORT_UNIX) == 0) ||
       (transport == vsock &&
-       (h->uri_allow_transports & LIBNBD_ALLOW_TRANSPORT_VSOCK) == 0)) {
+       (h->uri_allow_transports & LIBNBD_ALLOW_TRANSPORT_VSOCK) == 0) ||
+      (transport == ssh &&
+       (h->uri_allow_transports & LIBNBD_ALLOW_TRANSPORT_SSH) == 0)) {
     set_error (EPERM, "URI transport %s is not permitted", uri->scheme);
     goto cleanup;
   }
@@ -255,21 +303,29 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
 
   /* Parse the socket parameter. */
   for (i = 0; i < queries.len; i++) {
-    if (strcmp (queries.ptr[i].name, "socket") == 0)
+    if (strcasecmp (queries.ptr[i].name, "socket") == 0)
       unixsocket = queries.ptr[i].value;
   }
 
-  if (socket_required && !unixsocket) {
-    set_error (EINVAL, "cannot parse socket parameter from NBD URI "
-               "(did you mean to use \"%s:///?socket=...\"?)",
-               uri->scheme);
-    goto cleanup;
-  }
-  else if (!socket_required && unixsocket) {
-    set_error (EINVAL, "socket parameter is incompatible with \"%s:\" "
-               "(did you mean to use \"%s+unix:///?socket=...\"?)",
-               uri->scheme, !tls ? "nbd" : "nbds");
-    goto cleanup;
+  switch (socket_param) {
+  case required:
+    if (!unixsocket) {
+      set_error (EINVAL, "cannot parse socket parameter from NBD URI "
+                 "(did you mean to use \"%s:///?socket=...\"?)",
+                 uri->scheme);
+      goto cleanup;
+    }
+    break;
+  case none:
+    if (unixsocket) {
+      set_error (EINVAL, "socket parameter is incompatible with \"%s:\" "
+                 "(did you mean to use \"%s+unix:///?socket=...\"?)",
+                 uri->scheme, !tls ? "nbd" : "nbds");
+      goto cleanup;
+    }
+    break;
+  case optional:
+    /* no check needed */ ;
   }
 
   /* TLS */
@@ -278,7 +334,7 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
 
   /* Look for some tls-* parameters. */
   for (i = 0; i < queries.len; i++) {
-    if (strcmp (queries.ptr[i].name, "tls-certificates") == 0) {
+    if (strcasecmp (queries.ptr[i].name, "tls-certificates") == 0) {
       if (! h->uri_allow_local_file) {
         set_error (EPERM,
                    "local file access (tls-certificates) is not allowed, "
@@ -288,7 +344,7 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
       if (nbd_unlocked_set_tls_certificates (h, queries.ptr[i].value) == -1)
         goto cleanup;
     }
-    else if (strcmp (queries.ptr[i].name, "tls-psk-file") == 0) {
+    else if (strcasecmp (queries.ptr[i].name, "tls-psk-file") == 0) {
       if (! h->uri_allow_local_file) {
         set_error (EPERM,
                    "local file access (tls-psk-file) is not allowed, "
@@ -298,6 +354,17 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
       if (nbd_unlocked_set_tls_psk_file (h, queries.ptr[i].value) == -1)
         goto cleanup;
     }
+    else if (strcasecmp (queries.ptr[i].name, "tls-hostname") == 0) {
+      if (nbd_unlocked_set_tls_hostname (h, queries.ptr[i].value) == -1)
+        goto cleanup;
+    }
+    else if (strcasecmp (queries.ptr[i].name, "tls-verify-peer") == 0) {
+      int v = parse_bool ("tls-verify-peer", queries.ptr[i].value);
+      if (v == -1)
+        goto cleanup;
+      if (nbd_unlocked_set_tls_verify_peer (h, v) == -1)
+        goto cleanup;
+    }
   }
 
   /* Username. */
@@ -376,22 +443,49 @@ nbd_unlocked_aio_connect_uri (struct nbd_handle *h, const char *raw_uri)
      * uri->port > 0.  This prevents us from using certain very large
      * port numbers, but that's not an issue that matters in practice.
      */
-    svm_port = uri->port > 0 ? (uint32_t) uri->port : 10809;
+    svm_port = uri->port > 0 ? (uint32_t)uri->port : 10809;
     if (nbd_unlocked_aio_connect_vsock (h, cid, svm_port) == -1)
       goto cleanup;
 
     break;
   }
+
+  case ssh: {                   /* SSH */
+    char port_str[32];
+    const char *ssh_command[] = {
+      "ssh", "-p", port_str, uri->server,
+      "nc",
+      NULL,                     /* [5] "-U" or "localhost" */
+      NULL,                     /* [6] socket or "10809" */
+      NULL,
+    };
+
+    if (!uri->server || strcmp (uri->server, "") == 0) {
+      set_error (EINVAL, "SSH transport requires a server name");
+      goto cleanup;
+    }
+    snprintf (port_str, sizeof port_str,
+              "%d", uri->port > 0 ? uri->port : 22);
+
+    if (unixsocket) {
+      ssh_command[5] = "-U";
+      ssh_command[6] = unixsocket;
+    }
+    else {
+      ssh_command[5] = "localhost";
+      ssh_command[6] = "10809"; /* XXX provide a way to configure this */
+    }
+
+    if (nbd_unlocked_aio_connect_command (h, (char **) ssh_command) == -1)
+      goto cleanup;
+  }
   }
 
   ret = 0;
 
 cleanup:
-  for (i = 0; i < queries.len; ++i) {
-    free (queries.ptr[i].name);
-    free (queries.ptr[i].value);
-  }
-  free (queries.ptr);
+  uri_query_list_iter (&queries, free_uri_query);
+  uri_query_list_reset (&queries);
   xmlFreeURI (uri);
   return ret;
 }
@@ -422,7 +516,9 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
   else
     using_tls = false;
 
-  /* Set scheme, server or socket. */
+  /* First set uri.scheme and uri.server. */
+
+  /* We have a defined hostname and/or port number (nbd_connect_tcp). */
   if (h->hostname && h->port) {
     int r;
 
@@ -442,6 +538,7 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
     uri.server = server;
   }
 
+  /* We have a sockaddr (nbd_connect, nbd_connect_unix, nbd_connect_vsock). */
   else if (h->connaddrlen > 0) {
     switch (h->connaddr.ss_family) {
     case AF_INET:
@@ -451,7 +548,7 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
       char serv[NI_MAXSERV];
 
       uri.scheme = using_tls ? "nbds" : "nbd";
-      err = getnameinfo ((struct sockaddr *) &h->connaddr, h->connaddrlen,
+      err = getnameinfo ((struct sockaddr *)&h->connaddr, h->connaddrlen,
                          host, sizeof host, serv, sizeof serv,
                          NI_NUMERICHOST | NI_NUMERICSERV);
       if (err != 0) {
@@ -471,7 +568,7 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
     }
 
     case AF_UNIX: {
-      struct sockaddr_un *sun = (struct sockaddr_un *) &h->connaddr;
+      struct sockaddr_un *sun = (struct sockaddr_un *)&h->connaddr;
 
       if (sun->sun_path[0] == '\0') {
         /* Unix domain sockets in the abstract namespace are in theory
@@ -495,7 +592,7 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
 
 #if HAVE_STRUCT_SOCKADDR_VM
     case AF_VSOCK: {
-      struct sockaddr_vm *svm = (struct sockaddr_vm *) &h->connaddr;
+      struct sockaddr_vm *svm = (struct sockaddr_vm *)&h->connaddr;
 
       uri.scheme = using_tls ? "nbds+vsock" : "nbd+vsock";
       if (asprintf (&server, "%u:%u", svm->svm_cid, svm->svm_port) == -1) {
@@ -514,12 +611,13 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
     }
   }
 
+  /* Not enough information (nbd_connect_socket). */
   else {
     set_error (EINVAL, "cannot construct a URI for this connection type");
     goto out;
   }
 
-  /* Set other fields. */
+  /* Set other uri.* fields. */
   if (h->tls_username)
     uri.user = h->tls_username;
   if (h->export_name) {
@@ -539,11 +637,21 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
                              "tls-psk-file", h->tls_psk_file) == -1)
       goto out;
   }
+  if (h->tls_hostname) {
+    if (append_query_params (&query_params,
+                             "tls-hostname", h->tls_hostname) == -1)
+      goto out;
+  }
+  if (! h->tls_verify_peer) {
+    if (append_query_params (&query_params,
+                             "tls-verify-peer", "false") == -1)
+      goto out;
+  }
 
   uri.query_raw = query_params;
 
   /* Construct the final URI and return it. */
-  ret = (char *) xmlSaveUri (&uri);
+  ret = (char *)xmlSaveUri (&uri);
   if (ret == NULL)
     set_error (errno, "xmlSaveUri failed");
  out:
@@ -596,3 +704,28 @@ nbd_unlocked_get_uri (struct nbd_handle *h)
 }
 
 #endif /* !HAVE_LIBXML2 */
+
+/* nbd_is_uri is simple enough that it works even if we don't have URI
+ * support.
+ */
+int
+nbd_unlocked_is_uri (struct nbd_handle *h, const char *uri)
+{
+  static const char *prefixes[] = {
+    "nbd://",
+    "nbds://",
+    "nbd+unix://",
+    "nbds+unix://",
+    "nbd+vsock://",
+    "nbds+vsock://",
+    "nbd+ssh://",
+    "nbds+ssh://",
+  };
+  size_t i;
+
+  for (i = 0; i < ARRAY_SIZE (prefixes); ++i) {
+    if (strncasecmp (uri, prefixes[i], strlen (prefixes[i])) == 0)
+      return 1;
+  }
+  return 0;
+}
